"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RecoverMessageCommand = void 0;
const lisk_codec_1 = require("@liskhq/lisk-codec");
const lisk_tree_1 = require("@liskhq/lisk-tree");
const lisk_cryptography_1 = require("@liskhq/lisk-cryptography");
const lisk_chain_1 = require("@liskhq/lisk-chain");
const lisk_validator_1 = require("@liskhq/lisk-validator");
const types_1 = require("../../../../state_machine/types");
const base_interoperability_command_1 = require("../../base_interoperability_command");
const utils_1 = require("../../utils");
const schemas_1 = require("../../schemas");
const terminated_outbox_1 = require("../../stores/terminated_outbox");
const ccm_processed_1 = require("../../events/ccm_processed");
class RecoverMessageCommand extends base_interoperability_command_1.BaseInteroperabilityCommand {
    constructor() {
        super(...arguments);
        this.schema = schemas_1.messageRecoveryParamsSchema;
    }
    async verify(context) {
        const { params: { chainID, crossChainMessages, idxs, siblingHashes }, } = context;
        let terminatedOutboxAccount;
        try {
            lisk_validator_1.validator.validate(schemas_1.messageRecoveryParamsSchema, context.params);
        }
        catch (err) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: err,
            };
        }
        try {
            terminatedOutboxAccount = await this.stores.get(terminated_outbox_1.TerminatedOutboxStore).get(context, chainID);
        }
        catch (error) {
            if (error instanceof lisk_chain_1.NotFoundError) {
                return {
                    status: types_1.VerifyStatus.FAIL,
                    error: new Error('Terminated outbox account does not exist.'),
                };
            }
            throw error;
        }
        if (!crossChainMessages.length) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: new Error('No cross-chain messages to recover.'),
            };
        }
        if (idxs.length !== crossChainMessages.length) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: new Error('Inclusion proof indices and cross-chain messages do not have the same length.'),
            };
        }
        const sortedIdxs = [...idxs].sort((a, b) => a - b);
        const isSame = idxs.length === sortedIdxs.length &&
            idxs.every((element, index) => element === sortedIdxs[index]);
        if (!isSame) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: new Error('Cross-chain message indexes are not sorted in ascending order.'),
            };
        }
        if (idxs.length !== new Set(idxs).size) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: new Error('Cross-chain message indexes are not unique.'),
            };
        }
        if (idxs[0] < terminatedOutboxAccount.partnerChainInboxSize) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: new Error('Cross-chain message is not pending.'),
            };
        }
        for (const crossChainMessage of crossChainMessages) {
            const ccm = lisk_codec_1.codec.decode(schemas_1.ccmSchema, crossChainMessage);
            (0, utils_1.validateFormat)(ccm);
            if (ccm.status !== 0) {
                return {
                    status: types_1.VerifyStatus.FAIL,
                    error: new Error('Cross-chain message status is not valid.'),
                };
            }
            if (!ccm.receivingChainID.equals(chainID)) {
                return {
                    status: types_1.VerifyStatus.FAIL,
                    error: new Error('Cross-chain message receiving chain ID is not valid.'),
                };
            }
            const isLive = await this.internalMethod.isLive(context, ccm.sendingChainID, context.header.timestamp);
            if (!isLive) {
                return {
                    status: types_1.VerifyStatus.FAIL,
                    error: new Error('Cross-chain message sending chain is not live.'),
                };
            }
        }
        const proof = {
            size: terminatedOutboxAccount.outboxSize,
            idxs,
            siblingHashes,
        };
        const hashedCCMs = crossChainMessages.map(ccm => lisk_cryptography_1.utils.hash(ccm));
        const isVerified = lisk_tree_1.regularMerkleTree.verifyDataBlock(hashedCCMs, proof, terminatedOutboxAccount.outboxRoot);
        if (!isVerified) {
            return {
                status: types_1.VerifyStatus.FAIL,
                error: new Error('Message recovery proof of inclusion is not valid.'),
            };
        }
        return {
            status: types_1.VerifyStatus.OK,
        };
    }
    async execute(context) {
        const { params } = context;
        const recoveredCCMs = [];
        for (const crossChainMessage of params.crossChainMessages) {
            const ccmID = lisk_cryptography_1.utils.hash(crossChainMessage);
            const ccm = lisk_codec_1.codec.decode(schemas_1.ccmSchema, crossChainMessage);
            const ctx = {
                ...context,
                ccm,
                eventQueue: context.eventQueue.getChildQueue(ccmID),
            };
            if (ccm.sendingChainID.equals((0, utils_1.getMainchainID)(context.chainID))) {
                await this._applyRecovery(ctx);
            }
            else {
                await this._forwardRecovery(ctx);
                recoveredCCMs.push(crossChainMessage);
            }
        }
        const terminatedOutboxSubstore = this.stores.get(terminated_outbox_1.TerminatedOutboxStore);
        const terminatedOutboxAccount = await terminatedOutboxSubstore.get(context, context.params.chainID);
        const proof = {
            size: terminatedOutboxAccount.outboxSize,
            indexes: params.idxs,
            siblingHashes: params.siblingHashes,
        };
        terminatedOutboxAccount.outboxRoot = lisk_tree_1.regularMerkleTree.calculateRootFromUpdateData(recoveredCCMs.map(ccm => lisk_cryptography_1.utils.hash(ccm)), proof);
        await terminatedOutboxSubstore.set(context, context.params.chainID, terminatedOutboxAccount);
    }
    async _applyRecovery(context) {
        const { logger } = context;
        const { ccmID } = (0, utils_1.getEncodedCCMAndID)(context.ccm);
        const recoveredCCM = {
            ...context.ccm,
            status: 5,
            sendingChainID: context.ccm.receivingChainID,
            receivingChainID: context.ccm.sendingChainID,
        };
        try {
            for (const [module, method] of this.interoperableCCMethods.entries()) {
                if (method.verifyCrossChainMessage) {
                    logger.debug({
                        moduleName: module,
                        commandName: recoveredCCM.crossChainCommand,
                        ccmID: ccmID.toString('hex'),
                    }, 'Execute verifyCrossChainMessage');
                    await method.verifyCrossChainMessage(context);
                }
            }
        }
        catch (error) {
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 8,
                result: 3,
                ccm: recoveredCCM,
            });
            return;
        }
        const commands = this.ccCommands.get(recoveredCCM.module);
        if (!commands) {
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 2,
                result: 3,
                ccm: recoveredCCM,
            });
            return;
        }
        const command = commands.find(com => com.name === recoveredCCM.crossChainCommand);
        if (!command) {
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 3,
                result: 3,
                ccm: recoveredCCM,
            });
            return;
        }
        if (command.verify) {
            try {
                await command.verify(context);
            }
            catch (error) {
                logger.debug({
                    err: error,
                    moduleName: recoveredCCM.module,
                    commandName: recoveredCCM.crossChainCommand,
                }, 'Fail to verify cross chain command.');
                this.events
                    .get(ccm_processed_1.CcmProcessedEvent)
                    .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                    code: 9,
                    result: 3,
                    ccm: recoveredCCM,
                });
                return;
            }
        }
        const baseEventSnapshotID = context.eventQueue.createSnapshot();
        const baseStateSnapshotID = context.stateStore.createSnapshot();
        try {
            for (const [module, method] of this.interoperableCCMethods.entries()) {
                if (method.beforeCrossChainCommandExecute) {
                    logger.debug({
                        moduleName: module,
                        commandName: recoveredCCM.crossChainCommand,
                        ccmID: ccmID.toString('hex'),
                    }, 'Execute beforeCrossChainCommandExecute');
                    await method.beforeCrossChainCommandExecute(context);
                }
            }
        }
        catch (error) {
            context.eventQueue.restoreSnapshot(baseEventSnapshotID);
            context.stateStore.restoreSnapshot(baseStateSnapshotID);
            logger.debug({
                err: error,
                moduleName: recoveredCCM.module,
                commandName: recoveredCCM.crossChainCommand,
            }, 'Fail to execute beforeCrossChainCommandExecute.');
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 10,
                result: 3,
                ccm: recoveredCCM,
            });
            return;
        }
        const execEventSnapshotID = context.eventQueue.createSnapshot();
        const execStateSnapshotID = context.stateStore.createSnapshot();
        try {
            const params = command.schema ? lisk_codec_1.codec.decode(command.schema, recoveredCCM.params) : {};
            await command.execute({ ...context, params });
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 0,
                result: 0,
                ccm: recoveredCCM,
            });
        }
        catch (error) {
            context.eventQueue.restoreSnapshot(execEventSnapshotID);
            context.stateStore.restoreSnapshot(execStateSnapshotID);
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 4,
                result: 3,
                ccm: recoveredCCM,
            });
        }
        try {
            for (const [module, method] of this.interoperableCCMethods.entries()) {
                if (method.afterCrossChainCommandExecute) {
                    logger.debug({
                        moduleName: module,
                        commandName: recoveredCCM.crossChainCommand,
                        ccmID: ccmID.toString('hex'),
                    }, 'Execute afterCrossChainCommandExecute');
                    await method.afterCrossChainCommandExecute(context);
                }
            }
        }
        catch (error) {
            context.eventQueue.restoreSnapshot(baseEventSnapshotID);
            context.stateStore.restoreSnapshot(baseStateSnapshotID);
            logger.debug({ err: error, moduleName: module, commandName: recoveredCCM.crossChainCommand }, 'Fail to execute afterCrossChainCommandExecute');
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 11,
                result: 3,
                ccm: recoveredCCM,
            });
        }
    }
    async _forwardRecovery(context) {
        const { logger } = context;
        const encodedCCM = lisk_codec_1.codec.encode(schemas_1.ccmSchema, context.ccm);
        const ccmID = lisk_cryptography_1.utils.hash(encodedCCM);
        const recoveredCCM = {
            ...context.ccm,
            status: 5,
            sendingChainID: context.ccm.receivingChainID,
            receivingChainID: context.ccm.sendingChainID,
        };
        try {
            for (const [module, method] of this.interoperableCCMethods.entries()) {
                if (method.verifyCrossChainMessage) {
                    logger.debug({
                        moduleName: module,
                        commandName: recoveredCCM.crossChainCommand,
                        ccmID: ccmID.toString('hex'),
                    }, 'Execute verifyCrossChainMessage');
                    await method.verifyCrossChainMessage(context);
                }
            }
        }
        catch (error) {
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 8,
                result: 3,
                ccm: recoveredCCM,
            });
            logger.debug({
                err: error,
                moduleName: recoveredCCM.module,
                commandName: recoveredCCM.crossChainCommand,
            }, 'Fail to execute verifyCrossChainMessage.');
            return;
        }
        const baseEventSnapshotID = context.eventQueue.createSnapshot();
        const baseStateSnapshotID = context.stateStore.createSnapshot();
        try {
            for (const [module, method] of this.interoperableCCMethods.entries()) {
                if (method.beforeCrossChainMessageForwarding) {
                    logger.debug({
                        moduleName: module,
                        commandName: recoveredCCM.crossChainCommand,
                        ccmID: ccmID.toString('hex'),
                    }, 'Execute beforeCrossChainMessageForwarding');
                    await method.beforeCrossChainMessageForwarding(context);
                }
            }
        }
        catch (error) {
            context.eventQueue.restoreSnapshot(baseEventSnapshotID);
            context.stateStore.restoreSnapshot(baseStateSnapshotID);
            logger.debug({
                err: error,
                moduleName: recoveredCCM.module,
                commandName: recoveredCCM.crossChainCommand,
            }, 'Fail to execute beforeCrossChainMessageForwarding.');
            this.events
                .get(ccm_processed_1.CcmProcessedEvent)
                .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
                code: 12,
                result: 3,
                ccm: recoveredCCM,
            });
            return;
        }
        await this.internalMethod.addToOutbox(context, recoveredCCM.receivingChainID, recoveredCCM);
        this.events
            .get(ccm_processed_1.CcmProcessedEvent)
            .log(context, recoveredCCM.sendingChainID, recoveredCCM.receivingChainID, {
            code: 0,
            result: 1,
            ccm: recoveredCCM,
        });
    }
}
exports.RecoverMessageCommand = RecoverMessageCommand;
//# sourceMappingURL=recover_message.js.map